;;; rdb/proto.lisp --- RDB protocol

;; Extends OBJ/DB protocol with RocksDB-specific generics.

;;; Code:
(in-package :rdb)

;; Most methods are intended to work with the RDB struct directly
(defgeneric put-kv (self kv)
  (:documentation "Insert a KeyVal object."))
(defgeneric put-key (self key val)
  (:documentation "Insert a KEY and VAL."))
(defgeneric put-key-ts (self key val ts)
  (:documentation "Insert a KEY and VAL with associated timestamp TS."))
(defgeneric get-key (self key &key)
  (:documentation "Get value of KEY."))
(defgeneric multi-get (self keys &key)
  (:documentation "Retrieve multiple KEYS from SELF."))
(defgeneric get-opt (self key)
  (:documentation "Get value of option KEY."))
(defgeneric set-opt (self key val &key)
  (:documentation "Set value of option KEY to VAL."))
(defgeneric push-opts (self)
  (:documentation "Push all options to internal sap."))
(defgeneric backfill-opts (self &key)
  (:documentation "Backfill opts from an alien."))
(defgeneric push-sap (self key)
  (:documentation "Push a value associated with KEY to the sap associated
with SELF. Typically used to send a value from one slot, to a foreign
handle stored in another slot of the same object."))
(defgeneric push-sap* (self)
  (:documentation "Implicitly push values to the sap associated with SELF."))
(defgeneric pull-sap (self key)
  (:documentation "Pull a foreign value identified by KEY from the sap associated with SELF."))
(defgeneric pull-sap* (self)
  (:documentation "Implicitly pull foreign values from the sap associated with SELF."))
(defgeneric push-cf (self cf)
  (:documentation "Push a column-family to a sap."))
(defgeneric insert-key (self key val &key)
  (:documentation "Insert KEY:VAL into SELF."))
(defgeneric insert-kv (self kv &key)
  (:documentation "Insert KV object into SELF."))
(defgeneric make-kv (key val)
  (:documentation "Make a KV object from KEY:VAL."))
(defgeneric delete-key (self key &key)
  (:documentation "Delete value associated with KEY from SELF."))
(defgeneric delete-key-ts (self key ts)
  (:documentation "Delete value associated with KEY and TS from SELF."))
(defgeneric delete-key-range (self start end &key)
  (:documentation "Delete values associates with keys between START and END from SELF."))
(defgeneric make-transaction (self &key)
  (:documentation "Make a new transaction object from SELF."))
(defgeneric begin-transaction (self &key)
  (:documentation "Begin processing of transaction SELF."))
(defgeneric prepare-transaction (self &key)
  (:documentation "Prepare transaction SELF."))
(defgeneric rollback-transaction (self &key)
  (:documentation "Rollback transaction SELF."))
(defgeneric delete-transaction (self)
  (:documentation "Delete transaction SELF."))
(defgeneric commit-transaction (self &key)
  (:documentation "Commit transaction object SELF."))
(defgeneric flush-db (self &key)
  (:documentation "Flush the database SELF."))
(defgeneric sync-db (self other &key) ;;nyi
  (:documentation "Perform a synchronization on SELF using OTHER."))
(defgeneric find-cf (cf self &key)
  (:documentation "Find the column-familiy CF in SELF."))
(defgeneric flush-cf (self cf &key)
  (:documentation "Flush the column-family CF in SELF."))
(defgeneric repair-db (self &key)
  (:documentation "Attempt to repair the database SELF."))
(defgeneric backup-db (self &key)
  (:documentation "Create a new backup for database SELF."))
(defgeneric restore-db (self from &key)
  (:documentation "Restore database SELF from object FROM."))
(defgeneric snapshot-db (self)
  (:documentation "Create a new snapshot for database SELF."))
(defgeneric write-db (self batch &key)
  (:documentation "Write BATCH to database SELF."))
(defgeneric shutdown-db (self &key)
  (:documentation "Shutdown database SELF."))
(defgeneric ingest-db (self file &key)
  (:documentation "Ingest an external file into the database"))
(defgeneric get-prop (self propname)
  (:documentation "Get the property-value of PROPNAME from SELF."))
(defgeneric get-stats (self &optional htype)
  (:documentation "Get stats modulo HTYPE from SELF."))
(defgeneric print-stats (self &optional stream)
  (:documentation "Print statistics data from SELF."))
(defgeneric get-metadata (self &optional arg)
  (:documentation "Get metadata from SELF modulo ARG."))
(defgeneric create-iter (self &optional cf opts)
  (:documentation "Create an interator over the kvs of SELF module CF and OPTS."))
(defgeneric iter-next (self)
  (:documentation "Return the next value."))
(defgeneric iter-prev (self)
  (:documentation "Return the previous value."))
(defgeneric iter-seek (self key &key)
  (:documentation "Seek to a certain KEY in the iterator."))
(defgeneric iter-val (self)
  (:documentation "Return the value of current iterator item."))
(defgeneric iter-valid-p (self)
  (:documentation "Return non-nil if the iterator cursor is valid."))
(defgeneric iter-key (self)
  (:documentation "Return the key of current iterator item."))
(defgeneric iter-kv (self)
  (:documentation "Return the current KV object of the iterator by getting the key and
val."))
(defgeneric iter-timestamp (self)
  (:documentation "Return the timestamp of current iterator item."))
(defgeneric make-val (val)
  (:documentation "Coerce VAL into an OCTET-VECTOR.")
  (:method ((val null))
    #())
  (:method ((val string))
    (string-to-octets val))
  (:method ((val t))
    val))

(defgeneric make-key (key)
  (:documentation "Coerce KEY into an OCTET-VECTOR.")
  (:method ((val null))
    #())
  (:method ((val string))
    (string-to-octets val))
  (:method ((val t))
    val))

(defgeneric merge-key (self key val &key)
  (:documentation "Perform a merge operation on SELF using KEY and VAL."))

(defgeneric merge-kv (self kv &key)
  (:documentation "Perform a merge operation on SELF using object KV."))
